@@ -92,7 +92,7 @@
 #'   this is R's default behaviour. Use `FALSE` when specifically concerned
 #'   with the encoding, not just the value of the string.
 #' @param list_as_map Compare lists as if they are mappings between names and
-#'   values. Concretely, this drops `NULLs` in both objects and sorts named
+#'   values. Concretely, this drops `NULL`s in both objects and sorts named
 #'   components.
 #' @param quote_strings Should strings be surrounded by quotes? If `FALSE`,
 #'   only side-by-side and line-by-line comparisons will be used, and there's
@@ -124,18 +124,37 @@
 #' compare(list("x", "y"), list("x", "z"))
 #' compare(list(x = "x", x = "y"), list(x = "x", y = "z"))
 #'
-compare <- function(x, y, ...,
-                    x_arg = "old", y_arg = "new",
-                    tolerance = NULL,
-                    max_diffs = if (in_ci()) Inf else 10,
-                    ignore_srcref = TRUE,
-                    ignore_attr = "waldo_opts",
-                    ignore_encoding = TRUE,
-                    ignore_function_env = FALSE,
-                    ignore_formula_env = FALSE,
-                    list_as_map = FALSE,
-                    quote_strings = TRUE
-                    ) {
+compare <- function(
+  x,
+  y,
+  ...,
+  x_arg = "old",
+  y_arg = "new",
+  tolerance = NULL,
+  max_diffs = if (in_ci()) Inf else 10,
+  ignore_srcref = TRUE,
+  ignore_attr = "waldo_opts",
+  ignore_encoding = TRUE,
+  ignore_function_env = FALSE,
+  ignore_formula_env = FALSE,
+  list_as_map = FALSE,
+  quote_strings = TRUE
+) {
+  check_string(x_arg)
+  check_string(y_arg)
+  check_number_decimal(tolerance, allow_null = TRUE, min = 0)
+  check_number_whole(max_diffs, min = 1, allow_infinite = TRUE)
+  check_bool(ignore_srcref)
+  if (
+    !isTRUE(ignore_attr) && !isFALSE(ignore_attr) && !is.character(ignore_attr)
+  ) {
+    stop_input_type(ignore_attr, "a TRUE, a FALSE, or a character vector")
+  }
+  check_bool(ignore_encoding)
+  check_bool(ignore_function_env)
+  check_bool(ignore_formula_env)
+  check_bool(list_as_map)
+  check_bool(quote_strings)
 
   opts <- compare_opts(
     ...,
@@ -156,7 +175,12 @@ compare <- function(x, y, ...,
   new_compare(out, max_diffs)
 }
 
-compare_structure <- function(x, y, paths = c("x", "y"), opts = compare_opts()) {
+compare_structure <- function(
+  x,
+  y,
+  paths = c("x", "y"),
+  opts = compare_opts()
+) {
   if (!is_missing(x)) {
     proxy <- compare_proxy(x, paths[[1]])
     x <- proxy$object
@@ -168,7 +192,8 @@ compare_structure <- function(x, y, paths = c("x", "y"), opts = compare_opts())
     paths[[2]] <- proxy$path
   }
 
-  opts <- merge_lists(opts,
+  opts <- merge_lists(
+    opts,
     attr(x, "waldo_opts"),
     attr(y, "waldo_opts"),
     opts[opts$user_specified]
@@ -179,7 +204,10 @@ compare_structure <- function(x, y, paths = c("x", "y"), opts = compare_opts())
   }
 
   # Compare type
-  term <- compare_terminate(x, y, paths,
+  term <- compare_terminate(
+    x,
+    y,
+    paths,
     tolerance = opts$tolerance,
     ignore_attr = opts$ignore_attr
   )
@@ -201,14 +229,35 @@ compare_structure <- function(x, y, paths = c("x", "y"), opts = compare_opts())
 
   # Then attributes/slots
   if (isS4(x)) {
-    out <- c(out, compare_character(is(x), is(y), glue("is({paths})")))
+    out <- c(out, compare_character(is(x), is(y), glue::glue("is({paths})")))
     out <- c(out, compare_by_slot(x, y, paths, opts))
 
     # S4 objects can have attributes that are not slots
-    out <- c(out, compare_by_attr(
-      attrs(x, c(slotNames(x), "class")),
-      attrs(y, c(slotNames(y), "class")),
-      paths, opts)
+    out <- c(
+      out,
+      compare_by_attr(
+        attrs(x, c(slotNames(x), "class")),
+        attrs(y, c(slotNames(y), "class")),
+        paths,
+        opts
+      )
+    )
+  } else if (is.object(x) && inherits(x, "S7_object")) {
+    out <- c(
+      out,
+      compare_character(class(x), class(y), glue::glue("class({paths})"))
+    )
+    out <- c(out, compare_by_prop(x, y, paths, opts))
+
+    # S7 objects can have attributes that are not slots
+    out <- c(
+      out,
+      compare_by_attr(
+        attrs(x, c(S7::prop_names(x), "class", "S7_class")),
+        attrs(y, c(S7::prop_names(y), "class", "S7_class")),
+        paths,
+        opts
+      )
     )
   } else if (!isTRUE(opts$ignore_attr)) {
     if (is_call(x) && opts$ignore_formula_env) {
@@ -221,7 +270,19 @@ compare_structure <- function(x, y, paths = c("x", "y"), opts = compare_opts())
       y <- zap_srcref(y)
     }
 
-    out <- c(out, compare_by_attr(attrs(x, opts$ignore_attr), attrs(y, opts$ignore_attr), paths, opts))
+    if (compare_as_numeric(x, y, opts$tolerance)) {
+      opts$ignore_attr <- union(opts$ignore_attr, "class")
+    }
+
+    out <- c(
+      out,
+      compare_by_attr(
+        attrs(x, opts$ignore_attr),
+        attrs(y, opts$ignore_attr),
+        paths,
+        opts
+      )
+    )
   }
 
   # Then contents
@@ -239,7 +300,6 @@ compare_structure <- function(x, y, paths = c("x", "y"), opts = compare_opts())
     } else {
       out <- c(out, compare_by_pos(x, y, paths, opts))
     }
-
   } else if (is_environment(x)) {
     if (is_seen(list(x, y), opts$seen$envs)) {
       # Only report difference between pairs of environments once
@@ -252,8 +312,12 @@ compare_structure <- function(x, y, paths = c("x", "y"), opts = compare_opts())
       x_fields <- as.list.environment(x, all.names = TRUE)
       y_fields <- as.list.environment(y, all.names = TRUE)
       # Can't use as.list(sorted = TRUE), https://github.com/r-lib/waldo/issues/84
-      if (length(x_fields) > 0) x_fields <- x_fields[order(names(x_fields))]
-      if (length(y_fields) > 0) y_fields <- y_fields[order(names(y_fields))]
+      if (length(x_fields) > 0) {
+        x_fields <- x_fields[order(names(x_fields))]
+      }
+      if (length(y_fields) > 0) {
+        y_fields <- y_fields[order(names(y_fields))]
+      }
 
       if (env_has(x, ".__enclos_env__")) {
         # enclosing env of R6 methods is object env
@@ -264,8 +328,14 @@ compare_structure <- function(x, y, paths = c("x", "y"), opts = compare_opts())
 
       opts$seen$envs <- c(opts$seen$envs, list(list(x, y)))
       out <- c(out, compare_structure(x_fields, y_fields, paths, opts = opts))
-      out <- c(out, compare_structure(
-        parent.env(x), parent.env(y), paste0("parent.env(", paths, ")"), opts = opts)
+      out <- c(
+        out,
+        compare_structure(
+          parent.env(x),
+          parent.env(y),
+          paste0("parent.env(", paths, ")"),
+          opts = opts
+        )
       )
     }
   } else if (is_closure(x)) {
@@ -285,7 +355,9 @@ compare_structure <- function(x, y, paths = c("x", "y"), opts = compare_opts())
 
     if (!identical(x, y)) {
       diff <- compare_character(
-        deparse(x), deparse(y), paths,
+        deparse(x),
+        deparse(y),
+        paths,
         quote = "`",
         max_diffs = opts$max_diffs
       )
@@ -297,10 +369,15 @@ compare_structure <- function(x, y, paths = c("x", "y"), opts = compare_opts())
     }
   } else if (is_atomic(x)) {
     if (is_character(x) && !opts$ignore_encoding) {
-      out <- c(out, compare_character(
-        Encoding(x), Encoding(y), glue("Encoding({paths})"),
-        max_diffs = opts$max_diffs
-      ))
+      out <- c(
+        out,
+        compare_character(
+          Encoding(x),
+          Encoding(y),
+          glue::glue("Encoding({paths})"),
+          max_diffs = opts$max_diffs
+        )
+      )
     }
 
     out <- c(out, compare_vector(x, y, paths = paths, opts = opts))
@@ -316,8 +393,13 @@ compare_structure <- function(x, y, paths = c("x", "y"), opts = compare_opts())
     # Unevaluated dots are unlikely to lead to any significant differences
     # in behaviour (they're usually captured incidentally) so we just
     # ignore
+  } else if (typeof(x) == "weakref") {
+    out <- c(out, compare_by_weakref(x, y, paths, opts = opts))
   } else if (!typeof(x) %in% c("S4", "object")) {
-    abort(glue("{paths[[1]]}: unsupported type '{typeof(x)}'"), call = NULL)
+    abort(
+      glue::glue("{paths[[1]]}: unsupported type '{typeof(x)}'"),
+      call = NULL
+    )
   }
 
   out
@@ -351,9 +433,13 @@ is_identical <- function(x, y, opts) {
   }
 }
 
-compare_terminate <- function(x, y, paths,
-                              tolerance = NULL,
-                              ignore_attr = FALSE) {
+compare_terminate <- function(
+  x,
+  y,
+  paths,
+  tolerance = NULL,
+  ignore_attr = FALSE
+) {
   type_x <- friendly_type_of(x)
   type_y <- friendly_type_of(y)
   if (is_missing(x) && !is_missing(y)) {
@@ -365,23 +451,32 @@ compare_terminate <- function(x, y, paths,
     type_y <- col_c(type_y)
   }
 
-  type_mismatch_msg <- should_be("{type_x}{short_val(x)}", "{type_y}{short_val(y)}")
+  type_mismatch_msg <- should_be(
+    "{type_x}{short_val(x)}",
+    "{type_y}{short_val(y)}"
+  )
 
   # missing needs to be treated here because `typeof(missing_arg())` is symbol
   if (is_missing(x) != is_missing(y)) {
     return(type_mismatch_msg)
   }
 
-  if (typeof(x) == typeof(y) && oo_type(x) == oo_type(y)) {
-    return(character())
+  if (typeof(x) == typeof(y)) {
+    if (oo_type(x) == oo_type(y)) {
+      return(character())
+    }
+
+    if (oo_type(x) %in% c("base", "S3") && oo_type(y) %in% c("base", "S3")) {
+      return(character())
+    }
+
+    ignore_class <- isTRUE(ignore_attr) || "class" %in% ignore_attr
+    if (ignore_class) {
+      return(character())
+    }
   }
 
-  ignore_class <- isTRUE(ignore_attr) || "class" %in% ignore_attr
-  if (ignore_class && (typeof(x) == typeof(y))) {
-    return(character())
-  }
-
-  if (!is.null(tolerance) && is_numeric(x) && is_numeric(y)) {
+  if (compare_as_numeric(x, y, tolerance)) {
     return(character())
   }
 
@@ -395,10 +490,13 @@ compare_terminate <- function(x, y, paths,
 
 should_be <- function(x, y) {
   string <- paste0(
-    "`{paths[[1]]}` is ", x, "\n",
-    "`{paths[[2]]}` is ", y
+    "`{paths[[1]]}` is ",
+    x,
+    "\n",
+    "`{paths[[2]]}` is ",
+    y
   )
-  glue(string, .envir = caller_env(), .trim = FALSE)
+  glue::glue(string, .envir = caller_env(), .trim = FALSE)
 }
 
 # compare_each ------------------------------------------------------------
@@ -406,19 +504,23 @@ should_be <- function(x, y) {
 compare_by <- function(index_fun, extract_fun, path_fun) {
   function(x, y, paths, opts) {
     idx <- index_fun(x, y)
-    if (length(idx) == 0)
+    if (length(idx) == 0) {
       return(character())
+    }
 
     x_paths <- path_fun(paths[[1]], idx)
     y_paths <- path_fun(paths[[2]], idx)
 
     out <- character()
     for (i in seq_along(idx)) {
-      out <- c(out, compare_structure(
-        x = extract_fun(x, idx[[i]]),
-        y = extract_fun(y, idx[[i]]),
-        paths = c(x_paths[[i]], y_paths[[i]]),
-        opts = opts)
+      out <- c(
+        out,
+        compare_structure(
+          x = extract_fun(x, idx[[i]]),
+          y = extract_fun(y, idx[[i]]),
+          paths = c(x_paths[[i]], y_paths[[i]]),
+          opts = opts
+        )
       )
     }
 
@@ -427,37 +529,63 @@ compare_by <- function(index_fun, extract_fun, path_fun) {
 }
 
 index_name <- function(x, y) union(names(x), names(y))
-extract_name <- function(x, i) if (has_name(x, i)) .subset2(x, i) else missing_arg()
-path_name <- function(path, i) glue("{path}${i}")
+extract_name <- function(x, i) {
+  if (has_name(x, i)) .subset2(x, i) else missing_arg()
+}
+path_name <- function(path, i) glue::glue("{path}${i}")
 compare_by_name <- compare_by(index_name, extract_name, path_name)
 
 index_pos <- function(x, y) seq_len(max(length(x), length(y)))
-extract_pos <- function(x, i) if (i <= length(x)) .subset2(x, i) else missing_arg()
-path_pos <- function(path, i) glue("{path}[[{i}]]")
+extract_pos <- function(x, i) {
+  if (i <= length(x)) .subset2(x, i) else missing_arg()
+}
+path_pos <- function(path, i) glue::glue("{path}[[{i}]]")
 compare_by_pos <- compare_by(index_pos, extract_pos, path_pos)
 
-path_line <- function(path, i) glue("lines({path}[[{i}]])")
+path_line <- function(path, i) glue::glue("lines({path}[[{i}]])")
 compare_by_line <- compare_by(index_pos, extract_pos, path_line)
 
-path_line1 <- function(path, i) glue("lines({path})")
+path_line1 <- function(path, i) glue::glue("lines({path})")
 compare_by_line1 <- compare_by(index_pos, extract_pos, path_line1)
 
 path_attr <- function(path, i) {
   # from ?attributes, excluding row.names() because it's not a simple accessor
   funs <- c("comment", "class", "dim", "dimnames", "levels", "names", "tsp")
-  ifelse(i %in% funs, glue("{i}({path})"), glue("attr({path}, '{i}')"))
+  ifelse(
+    i %in% funs,
+    glue::glue("{i}({path})"),
+    glue::glue("attr({path}, '{i}')")
+  )
 }
 compare_by_attr <- compare_by(index_name, extract_name, path_attr)
 
 #' @importFrom methods slotNames .hasSlot slot is
 index_slot <- function(x, y) union(slotNames(x), slotNames(y))
 extract_slot <- function(x, i) if (.hasSlot(x, i)) slot(x, i) else missing_arg()
-path_slot <- function(path, i) glue("{path}@{i}")
+path_slot <- function(path, i) glue::glue("{path}@{i}")
 compare_by_slot <- compare_by(index_slot, extract_slot, path_slot)
 
+# We compare S7 based not on their properties (which might have getters/setters)
+# but instead the underlying attributes that store any data. This might generate
+# confusing messages (e.g. if obj@x <- 1 actually sets property y)
+index_prop <- function(x, y) union(attr_names(x), attr_names(y))
+extract_prop <- function(x, i) {
+  if (S7::prop_exists(x, i)) S7::prop(x, i) else missing_arg()
+}
+path_prop <- function(path, i) glue::glue("{path}@{i}")
+compare_by_prop <- compare_by(index_prop, extract_prop, path_prop)
+attr_names <- function(x) names(attributes(x))
 extract_fun <- function(x, i) switch(i, fn_body(x), fn_fmls(x), fn_env(x))
 path_fun <- function(path, i) {
   fun <- unname(c("body", "formals", "environment")[i])
-  glue("{fun}({path})")
+  glue::glue("{fun}({path})")
 }
 compare_by_fun <- compare_by(function(x, y) 1:3, extract_fun, path_fun)
+
+index_weakref <- function(x, y) 1:2
+extract_weakref <- function(x, i) switch(i, wref_key(x), wref_value(x))
+path_weakref <- function(path, i) {
+  fun <- unname(c("weakref_key", "weakref_value")[i])
+  glue::glue("{fun}({path})")
+}
+compare_by_weakref <- compare_by(index_weakref, extract_weakref, path_weakref)
